{ *****************************************************
               TNavigator Component

     A set of generic navigator button bars.

                  Paul Warren
         HomeGrown Software Development
       (c) 1997 Langley British Columbia.
                (604) 856-6523
         e-mail:  hg_soft@uniserve.com
    Home page: http://users.uniserve.com/~hg_soft
  ***************************************************** }

unit Navigate;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, ExtCtrls, Buttons, DsgnIntf, EnhCalnd;

type
  TNavigatorClass = class of TNavigator;
  TNavigatorButton = class;

  TNavClick = procedure(Sender: TObject; Button: integer) of object;

  TCustomNavigator = class(TCustomPanel)
  private
    { Private declarations }
    FButtons: TStringList;
    FGlyphs: TBitMap;
    FHints: TStrings;
    FOnNavClick: TNavClick;
    FUseCaptions: boolean;
    ButtonWidth: Integer;
    ButtonHeight: Integer;
    MinBtnSize: TPoint;
    FocusedButton: integer;
    procedure BtnMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer); virtual;
    procedure InitButtons; virtual;
    procedure ClearButtons;
    procedure InitHints;
    procedure SetHints(Value: TStrings);
    procedure SetButtons(Value: TStringList);
    procedure SetGlyphs(Value: TBitMap);
    procedure SetUseCaptions(Value: boolean);
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
    procedure WMKillFocus(var Message: TWMKillFocus); message WM_KILLFOCUS;
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
  protected
    { Protected declarations }
    procedure Loaded; override;
    procedure BtnClick(Sender: TObject); virtual;
    function GetBtnGlyph(AIndex: integer): TBitMap;
    procedure AdjustSize(var W: Integer; var H: Integer);
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
    property Buttons: TStringList read FButtons write SetButtons;
    property Glyphs: TBitMap read FGlyphs write SetGlyphs;
    property Hints: TStrings read FHints write SetHints;
    property UseCaptions: boolean read FUseCaptions write SetUseCaptions default false;
    property OnBtnClick: TNavClick read FOnNavClick write FOnNavClick;
  published
    { Published declarations }
  end;

  TNavigatorButton = class(TSpeedButton)
  private
    { Private declarations }
    FIndex: integer;
  protected
    { protected declarations }
    procedure Paint; override;
  public
    { public declarations }
    property Index: integer read FIndex write FIndex;
  published
    { published declarations }
  end;

  TNavigator = class(TCustomNavigator)
  private
    { Private declarations }
  protected
    { protected declarations }
  public
    { public declarations }
  published
    { published declarations }
    property Buttons;
    property Glyphs;
    property Hints;
    property UseCaptions;
    property OnBtnClick;
    property Align;
    property Ctl3D;
    property Enabled;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnEnter;
    property OnExit;
  end;

  TCalendarNav = class(TCustomNavigator)
  private
    { Private declarations }
    FCalendarSource: TBaseCalendar;
    procedure SetSource(Value: TBaseCalendar);
  protected
    { protected declarations }
    procedure BtnClick(Sender: TObject); override;
    procedure Notification(AComponent: TComponent;
      Operation: TOperation); override;
  public
    { public declarations }
    constructor Create(AOwner: TComponent); override;
  published
    { published declarations }
    property CalendarSource: TBaseCalendar read FCalendarSource write SetSource;
    property Buttons;
    property Glyphs;
    property Hints;
    property UseCaptions;
    property Align;
    property Ctl3D;
    property Enabled;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnEnter;
    property OnExit;
  end;

const
  VerbVisual = 'Navigator Bar Editor';

type
  { Component Editor for TNavigator }
  TNavigatorEditor = class(TComponentEditor)
    procedure Edit; override;
    procedure ExecuteVerb(Index: Integer); override;
    function GetVerb(Index: Integer): string; override;
    function GetVerbCount: Integer; override;
  end;

procedure Register;

implementation

uses  NavEdit;

{$IFDEF WIN32}
  {$R NAVIGATE.D32}
  {$R NAV32.RES}
{$ELSE}
  {$R NAVIGATE.D16}
  {$R NAV16.RES}
{$ENDIF}

{ TNavigator }
constructor TCustomNavigator.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle - [csAcceptsControls, csSetCaption ] + [csFramed, csOpaque];
  { set dimensions }
  Width := 160;
  Height := 32;
  { defaults }
  BevelOuter := bvNone;
  BevelInner := bvNone;
  ButtonWidth := 0;
  ButtonHeight := 0;
  FocusedButton := 0;
  TabStop := false;
  FGlyphs := TBitMap.Create;
  FUseCaptions := false;
  FButtons := TStringList.Create;
  { create default buttons }
  FButtons.Add('BTPREV');
  FButtons.Add('BTNEXT');
  { initialize default buttons }
  InitButtons;
  FHints := TStringList.Create;
  { create default hints }
  FHints.Add('Go Back');
  FHints.Add('Go Forward');
  { initialize default hints }
  InitHints;
end;

destructor TCustomNavigator.Destroy;
begin
  ClearButtons;
  FGlyphs.Free;
  FHints.Free;
  FButtons.Free;
  inherited Destroy;
end;

procedure TCustomNavigator.Loaded;
begin
  inherited Loaded;
  { clear old buttons first }
  ClearButtons;
  { initialize any non-default buttons }
  InitButtons;
  { initialize any non-default hints }
  InitHints;
end;

procedure TCustomNavigator.ClearButtons;
var
  i: integer;
begin
  { free all buttons }
  for i := 0 to FButtons.Count-1 do
    FButtons.Objects[i].Free;
end;

procedure TCustomNavigator.InitButtons;
var
  Btn: TNavigatorButton;
  i, x: integer;
  S: array[0..79] of char;
begin
  MinBtnSize := Point(20, 18);
  X := 0;
  { initiallize buttons, one to each TStringList entry }
  for I := 0 to FButtons.Count-1 do
  begin
    { create a button }
    Btn := TNavigatorButton.Create(Self);
    { set properties }
    Btn.Index := i;
    Btn.Enabled := Enabled;
    if FGlyphs.Empty then
      Btn.Glyph.Handle := LoadBitmap(HInstance, StrPCopy(S, FButtons.Strings[i]))
    else
      Btn.Glyph := GetBtnGlyph(i);
    Btn.SetBounds(X, 0, MinBtnSize.X, MinBtnSize.Y);
    Btn.Visible := true;
    if FUseCaptions then Btn.Caption := FButtons.Strings[i];
    Btn.OnClick := BtnClick;
    Btn.OnMouseDown := BtnMouseDown;
    Btn.Parent := Self;
    { add button to string list object }
    FButtons.Objects[i] := Btn;
    X := X + MinBtnSize.X;
  end;
  SetBounds(Left, Top, Width, Height);
end;

function TCustomNavigator.GetBtnGlyph(AIndex: integer): TBitMap;
var
  Temp: TBitMap;
  W: integer;
begin
  { calc width }
  W := FGlyphs.Width div FButtons.Count;
  { create temp bitmap }
  Temp := TBitMap.Create;
  { set width and height }
  Temp.Height := FGlyphs.Height;
  Temp.Width := W;
  { copy appropriate piece of bitmap strip to Temp }
  Temp.Canvas.CopyRect(Rect(0, 0, W, FGlyphs.Height),
    FGlyphs.Canvas, Rect(AIndex*W, 0, (AIndex*W)+W, FGlyphs.Height));
  { set result equal to temp, don't use Assign
  we don't want a copy }
  Result := Temp;
end;

procedure TCustomNavigator.InitHints;
var
  i: integer;
begin
  for i := 0 to (FHints.Count - 1) do
    if (FHints.Strings[i] <> '') and (i <= FButtons.Count-1) then
        (FButtons.Objects[i] as TNavigatorButton).Hint := FHints.Strings[i];
end;

procedure TCustomNavigator.SetButtons(Value: TStringList);
begin
  { clear old buttons first }
  ClearButtons;
  { set new buttons }
  FButtons.Assign(Value);
  { initialize }
  InitButtons;
end;

procedure TCustomNavigator.SetGlyphs(Value: TBitMap);
begin
  { clear old buttons first }
  ClearButtons;
  { set new glyphs }
  FGlyphs.Assign(Value);
  { initialize }
  InitButtons;
end;

procedure TCustomNavigator.SetUseCaptions(Value: boolean);
begin
  if Value <> FUseCaptions then
  begin
    FUseCaptions := Value;
    { clear old buttons first }
    ClearButtons;
    { initialize }
    InitButtons;
  end;
end;

procedure TCustomNavigator.SetHints(Value: TStrings);
begin
  FHints.Assign(Value);
  InitHints;
end;

procedure TCustomNavigator.AdjustSize(var W: Integer; var H: Integer);
var
  MinW: Integer;
  MinH: Integer;
  Space, Temp, Remain: Integer;
  i, X, Y: Integer;
begin
  if (csLoading in ComponentState) then Exit;
  if FButtons = nil then Exit;
  if FButtons.Objects[0] = nil then Exit;

  if W >= H then
  begin
    MinW := FButtons.Count * (MinBtnSize.X - 1) + 1;
    if W < MinW then
      W := MinW;
    if H < MinBtnSize.Y then
      H := MinBtnSize.Y;

    ButtonWidth := ((W - 1) div FButtons.Count) + 1;
    Temp := FButtons.Count * (ButtonWidth - 1) + 1;
    if Align = alNone then
      W := Temp;

    X := 0;
    Remain := W - Temp;
    Temp := FButtons.Count div 2;
    for I := 0 to FButtons.Count-1 do
    begin
      Space := 0;
      if Remain <> 0 then
      begin
        Dec(Temp, Remain);
        if Temp < 0 then
        begin
          Inc(Temp, FButtons.Count);
          Space := 1;
        end;
      end;
      (FButtons.Objects[I] as TNavigatorButton).SetBounds(X, 0, ButtonWidth + Space, Height);
      Inc(X, ButtonWidth - 1 + Space);
    end;
  end else
  begin
    MinH := FButtons.Count * (MinBtnSize.Y - 1) + 1;
    if H < MinH then
      H := MinH;
    if W < MinBtnSize.X then
      W := MinBtnSize.X;

    ButtonHeight := ((H - 1) div FButtons.Count) + 1;
    Temp := FButtons.Count * (ButtonHeight - 1) + 1;
    if Align = alNone then
      H := Temp;

    Y := 0;
    Remain := H - Temp;
    Temp := FButtons.Count div 2;
    for I := 0 to FButtons.Count-1 do
    begin
      Space := 0;
      if Remain <> 0 then
      begin
        Dec(Temp, Remain);
        if Temp < 0 then
        begin
          Inc(Temp, FButtons.Count);
          Space := 1;
        end;
      end;
      (FButtons.Objects[I] as TNavigatorButton).SetBounds(0, Y, Width, ButtonHeight + Space);
      Inc(Y, ButtonHeight - 1 + Space);
    end;
  end;
end;

procedure TCustomNavigator.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
var
  W, H: Integer;
begin
  W := AWidth;
  H := AHeight;
  AdjustSize(W, H);
  inherited SetBounds(ALeft, ATop, W, H);
end;

procedure TCustomNavigator.KeyDown(var Key: Word; Shift: TShiftState);
var
  NewFocus: integer;
  OldFocus: integer;
begin
  OldFocus := FocusedButton;
  case Key of
    VK_RIGHT:
      begin
        NewFocus := FocusedButton;
        repeat
          if NewFocus < FButtons.Count-1 then
            NewFocus := Succ(NewFocus);
        until (NewFocus = FButtons.Count-1) or ((FButtons.Objects[NewFocus] as TNavigatorButton).Visible);
        if NewFocus <> FocusedButton then
        begin
          FocusedButton := NewFocus;
          (FButtons.Objects[OldFocus] as TNavigatorButton).Invalidate;
          (FButtons.Objects[FocusedButton] as TNavigatorButton).Invalidate;
        end;
      end;
    VK_LEFT:
      begin
        NewFocus := FocusedButton;
        repeat
          if NewFocus > 0 then
            NewFocus := Pred(NewFocus);
        until (NewFocus = 0) or ((FButtons.Objects[NewFocus] as TNavigatorButton).Visible);
        if NewFocus <> FocusedButton then
        begin
          FocusedButton := NewFocus;
          (FButtons.Objects[OldFocus] as TNavigatorButton).Invalidate;
          (FButtons.Objects[FocusedButton] as TNavigatorButton).Invalidate;
        end;
      end;
    VK_SPACE: if (FButtons.Objects[FocusedButton] as TNavigatorButton).Enabled then
                (FButtons.Objects[FocusedButton] as TNavigatorButton).Click;
  end;
end;

procedure TCustomNavigator.WMSize(var Message: TWMSize);
begin
  inherited;
  SetBounds(Left, Top, Width, Height);
  Message.Result := 0;
end;

procedure TCustomNavigator.WMSetFocus(var Message: TWMSetFocus);
begin
  if not (csLoading in ComponentState) then
    (FButtons.Objects[FocusedButton] as TNavigatorButton).Invalidate;
end;

procedure TCustomNavigator.WMKillFocus(var Message: TWMKillFocus);
begin
  if not (csLoading in ComponentState) then
    (FButtons.Objects[FocusedButton] as TNavigatorButton).Invalidate;
end;

procedure TCustomNavigator.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  Message.Result := DLGC_WANTARROWS;
end;

procedure TCustomNavigator.BtnClick(Sender: TObject);
begin
  if not (csDesigning in ComponentState) and Assigned(FOnNavClick)
    and (Sender as TNavigatorButton).Enabled then
      FOnNavClick(Self, TNavigatorButton(Sender).Index);
end;

procedure TCustomNavigator.BtnMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  OldFocus: integer;
begin
  OldFocus := FocusedButton;
  FocusedButton := TNavigatorButton(Sender).Index;
  if TabStop and (GetFocus <> Handle) and CanFocus then
  begin
    SetFocus;
    if (GetFocus <> Handle) then
      Exit;
  end
  else if TabStop and (GetFocus = Handle) and (OldFocus <> FocusedButton) then
  begin
    (FButtons.Objects[OldFocus] as TNavigatorButton).Invalidate;
    (FButtons.Objects[FocusedButton] as TNavigatorButton).Invalidate;
  end;
end;

{ TNavigatorButton }
procedure TNavigatorButton.Paint;
var
  R: TRect;
begin
  inherited Paint;
  if (GetFocus = Parent.Handle) and
    (FIndex = TNavigator(Parent).FocusedButton) then
  begin
    R := Bounds(0, 0, Width, Height);
    InflateRect(R, -3, -3);
    if FState = bsDown then OffsetRect(R, 1, 1);
    DrawFocusRect(Canvas.Handle, R);
  end;
end;

{ TCalendarNav }
constructor TCalendarNav.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  { load Bitmap from resource }
  FGlyphs.Handle := LoadBitmap(HInstance, 'CALNAV');
  ClearButtons;
  FButtons.Clear;
  { create new default buttons }
  FButtons.Add('PrevY');
  FButtons.Add('Prev');
  FButtons.Add('Next');
  FButtons.Add('NextY');
  { initialize new default buttons }
  InitButtons;
  { clear existing default hints }
  FHints.Clear;
  { create new default hints }
  FHints.Add('Previous Year');
  FHints.Add('Previous Month');
  FHints.Add('Next Month');
  FHints.Add('Next Year');
  { initialize new default hints }
  InitHints;
end;

procedure TCalendarNav.SetSource(Value: TBaseCalendar);
begin
  { set FCalendarSource := Value }
  FCalendarSource := Value;
end;

procedure TCalendarNav.Notification(AComponent: TComponent;
      Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  { If the connected TBaseCalendar has been deleted, make the connection nil }
  if (Operation = opRemove) and (AComponent = FCalendarSource) then
     FCalendarSource := nil;
end;

procedure TCalendarNav.BtnClick(Sender: TObject);
begin
  if FCalendarSource <> nil then
    case (Sender as TNavigatorButton).Index of
      0: FCalendarSource.PrevYear;
      1: FCalendarSource.PrevMonth;
      2: FCalendarSource.NextMonth;
      3: FCalendarSource.NextYear;
    end;
end;

{ TNavigator Component Editor }
procedure TNavigatorEditor.Edit;
var
  NavEditForm: TNavEditForm;
begin
  NavEditForm := TNavEditForm.Create(Application);
  if Component is TNavigator then with (Component as TNavigator) do
  begin
    NavEditForm.SpinEdit1.Value := Buttons.Count;
    NavEditForm.Memo1.Lines.Assign(Buttons);
    NavEditForm.Memo2.Lines.Assign(Hints);
    NavEditForm.Navigator1.Glyphs.Assign(Glyphs);
    NavEditForm.CheckBox1.Checked := UseCaptions;
    NavEditForm.CheckBox2.Checked := ShowHint;
  end;
  try
    { Show the property editor Modal }
    if NavEditForm.ShowModal = mrOk then
      if Component is TNavigator then with (Component as TNavigator) do
      begin
        Glyphs.Assign(NavEditForm.Navigator1.Glyphs);
        UseCaptions := NavEditForm.Navigator1.UseCaptions;
        ShowHint := NavEditForm.Navigator1.ShowHint;
        Hints := TStringList(NavEditForm.Memo2.Lines);
        Buttons := TStringList(NavEditForm.Memo1.Lines);
        Designer.Modified;
      end;
  finally
    NavEditForm.Release;
  end;
end;

procedure TNavigatorEditor.ExecuteVerb(Index: Integer);
begin
  if Index = 0 then Edit;
end;

function TNavigatorEditor.GetVerb(Index: Integer): string;
begin
  Result := VerbVisual;
end;

function TNavigatorEditor.GetVerbCount: Integer;
begin
  Result := 1;
end;

{ register component on Misc page }
procedure Register;
begin
  RegisterComponents('HomeGrown', [TNavigator]);
  RegisterComponents('HomeGrown', [TCalendarNav]);
  RegisterComponentEditor(TNavigator, TNavigatorEditor);
  RegisterComponentEditor(TCalendarNav, TNavigatorEditor);
end;

end.